# coding: utf-8
"""
Description:
    Internal signal callbacks allow:
    --> Field to create Item
    --> Item to create Translation
    --> Language to create Translation
    External signal callbacks allow any declared table to create or delete related "Item" instance
    And those "Item" instances then create "Translation" instances using or internal signal callbacks
Signal Internal Callbacks:
    create_items_from_field: Creates a new Item instance for this field, for every existing object of the model's field
    create_translations_from_item: Creates Translation instances with our item for each available language
    create_translations_from_language: Creates new Translation entry for every unique "item" in Translation
Signal External Callbacks:
    create_translated_items: Creates Item instances everytime an object is created in a translated table
    delete_translated_items: Deletes Item instances everytime an object is deleted in a translated table
Applying the External Callbacks:
    This snippet allows us to get the models within the app that herit from our "TranslationModel"
    And then applies the external callbacks to those applications
"""


# --------------------------------------------------------------------------------
# > Imports
# --------------------------------------------------------------------------------
# Built-in

# Django
from django.apps import apps
from django.db.models.signals import post_save, post_delete
from django.dispatch import receiver

# Third-party

# Local
from .models import Field, Item, Language, Translation


# --------------------------------------------------------------------------------
# > Signal Internal Callbacks
# --------------------------------------------------------------------------------
@receiver(post_save, sender=Field)
def create_items_from_field(sender, instance, created, **kwargs):
    """
    Creates a new Item instance for this field, for every existing object of the model's field
    Technically, this callback does several things:
    - It gets all the existing objects from the model where our Field supposedly comes from
    - For each object, it creates the associated Item (Field + Object)
    - Then it adds the new Item PK as a FK back into the object instance
    Since "bulk_create" cannot be used on Item, so we loop to create them individually
    """
    if created:
        # Get the class model associated with the new Field
        app_name = instance.content_type.app_label
        model_name = instance.content_type.model
        target_model = apps.get_model(app_name, model_name)
        # For each object, creates a new item and adds the FK back into the object instance
        objects = target_model.objects.all()
        for obj in objects:
            item = Item.objects.create(field=instance, object_id=obj.pk, content_type=instance.content_type)
            setattr(obj, instance.name, item)
            obj.save()


@receiver(post_save, sender=Item)
def create_translations_from_item(sender, instance, created, **kwargs):
    """Creates Translation instances in every Language for our new Item"""
    if created:
        languages = Language.objects.all()
        if len(languages) > 0:
            translations = [Translation(item=instance, language=language) for language in languages]
            Translation.objects.bulk_create(translations)


@receiver(post_save, sender=Language)
def create_translations_from_language(sender, instance, created, **kwargs):
    """Creates Translation for our new Language and all existing Item instances"""
    if created:
        items = Item.objects.all()
        if len(items) > 0:
            translations = [Translation(language=instance, item=item) for item in items]
            Translation.objects.bulk_create(translations)


# --------------------------------------------------------------------------------
# > Signal External Callbacks
# --------------------------------------------------------------------------------
def create_translated_items(sender, instance, created, **kwargs):
    """
    Creates Item instances everytime an object is created in a translated table. Note that:
    - "bulk_create" cannot be used on Item, so we use a loop
    - we use "setattr" because we have the field name, not the Field instance
    """
    if created:
        fields = instance.get_translated_fields()
        if len(fields) > 0:
            for field in fields:
                content_type = instance.get_content_type_instance()
                item = Item.objects.create(field=field, object_id=instance.pk, content_type=content_type)
                setattr(instance, field.name, item)
                instance.save()
        else:
            raise RuntimeError("{} has no entry in the Field table".format(sender))


def delete_translated_items(sender, instance, **kwargs):
    """
    Deletes Item instances everytime an object is deleted in a translated table
    Then "Translations" are automatically deleted due to its CASCADE relationship with "Item"
    """
    fields = instance.get_translated_fields()
    if len(fields) > 0:
        # We get the FK towards the Item model and filter the missing key to avoid errors
        items = [getattr(instance, field.name) for field in fields]
        filtered_items = [x for x in items if x is not None]
        if len(filtered_items) > 0:
            # We have Item instances, but only want their "id" attribute
            id_list = map(lambda x: x.id, filtered_items)
            Item.objects.filter(pk__in=id_list).delete()


# --------------------------------------------------------------------------------
# > Applying the External Callbacks
# --------------------------------------------------------------------------------
# Getting the models subjects to translation
translated_models = []
for model in apps.get_models():
    try:
        if model.TRANSLATED:
            translated_models.append(model)
    except AttributeError:
        pass

# Applying our signals to those models:
for model in translated_models:
    receiver(post_save, sender=model)(create_translated_items)
    receiver(post_delete, sender=model)(delete_translated_items)



# coding: utf-8
"""
Description:
    Contains helper functions to assist you when working with database translations
Functions:
    all_instances_as_translated_dict: Applies 'instance_as_translated_dict' to the iterable of instances
    get_current_language: Returns the current active language. Will set a default language if none is found.
    instance_as_translated_dict: Returns a model instance into a dict containing all of its fields
    set_default_language: Sets the default language if none is chosen
    update_user_language: Updates the user current language following Django guildelines
"""


# --------------------------------------------------------------------------------
# > Imports
# --------------------------------------------------------------------------------
# Built-in

# Django
from django.db import models
from django.db.models.fields.files import ImageFieldFile, FieldFile
from django.utils.translation import activate, LANGUAGE_SESSION_KEY

# Third-party

# Local
from .models import Item, Language, Translation


# --------------------------------------------------------------------------------
# > Functions
# --------------------------------------------------------------------------------
def all_instances_as_translated_dict(instances, depth=True, language=None, request=None):
    """
    Description:
        Applies 'instance_as_translated_dict' to the iterable of instances
        Returns a list of dicts which contains the fields of all your instances
        Check the 'instance_as_translated_dict' for more info
    Args:
        instances (iterable): An iterable of your model instances
        depth (bool, optional): Determines if FK will also be transformed into dicts. Defaults to True.
        language (Language, optional): A Language instance from this app. Defaults to None.
        request (HttpRequest, option): HttpRequest from Django. Defaults to None.
    Returns:
        list: A list of dicts, where each dict contains the fields/values of the initial instances
    """
    # Checking arguments
    if language is None and request is None:
        raise TypeError("You must provide either 'language' or 'request'")
    # Get the language from the session
    if language is None:
        language = get_current_language(request)
    # Loop over instances
    results = []
    for instance in instances:
        result = instance_as_translated_dict(instance, depth=depth, language=language)
        results.append(result)
    return results


def get_current_language(request, set_default=True, default_id=1):
    """
    Description:
        Returns the current active language. Will set a default language if none is found.
    Args:
        request (HttpRequest): HttpRequest from Django
        set_default (Boolean): Indicates if a default language must be activated (if none currently is). Default to True.
        default_id (Integer): The PK for the default Language instance. Default to 1
    Returns:
        Language: The currently used language from our app's Language model
    """
    # Base variables
    language = None
    language_name = request.session.get(LANGUAGE_SESSION_KEY, False)
    # Get the language
    if language_name:
        try:
            language = Language.objects.get(django_language_name=language_name)
        except Language.DoesNotExist:
            pass
    # Set a default language if necessary
    if language is None and set_default:
        language = set_default_language(request, default_id)
    # Always return the active language
    return language


def instance_as_translated_dict(instance, depth=True, language=None, request=None):
    """
    Description:
        Returns a model instance into a dict containing all of its fields
        Language can be given as an argument, or guess through the user of "request"
        With "depth" set to True, ForeignKey will also be transformed into sub-dict
        Files and images are replaced by a subdict with 'path', 'url', and 'name' keys
        Meaning you will be able to manipulate the dict in an HTML template much like an instance
    Args:
        instance (Model): An instance from any of your models
        depth (bool, optional): Determines if FK will also be transformed into dicts. Defaults to True.
        language (Language, optional): A Language instance from this app. Defaults to None.
        request (HttpRequest, option): HttpRequest from Django. Defaults to None.
    Returns:
        dict: A dict with all of the instance's fields and values
    """
    # Checking arguments
    if language is None and request is None:
        raise TypeError("You must provide either 'language' or 'request'")
    # Get the language from the session
    if language is None:
        language = get_current_language(request)
    # Loop over fields
    translated_dict = {}
    fields = instance._meta.get_fields()
    for field in fields:
        value = getattr(instance, field.name, None)
        if value is not None:
            value_type = type(value)
            # Case 1: Get the translation
            if value_type == Item:
                new_value = Translation.objects.get(item=value, language=language).text
            # Case 2: Go to the linked model and repeat the process (unless depth=False)
            elif issubclass(value_type, models.Model):
                if depth:
                    new_value = instance_as_translated_dict(value, depth=True, language=language)
                else:
                    new_value = value
            # Case 3:
            elif value_type in {ImageFieldFile, FieldFile}:
                if value:
                    new_value = {
                        "name": getattr(value, "name", ""),
                        "url": getattr(value, "url", ""),
                        "path": getattr(value, "path", ""),
                    }
                else:
                    new_value = ""
            # Case 4: Keep the value as it is
            else:
                new_value = value
            translated_dict[field.name] = new_value
    return translated_dict


def set_default_language(request, pk=1):
    """Sets the default language if none is chosen"""
    language = Language.objects.get(id=pk)
    update_user_language(request, language=language)
    return language


def update_user_language(request, language=None, language_id=None):
    """
    Description:
        Updates the user current language following Django guildelines
        This will allow for both "Django" frontend translations and "our app" database translation
        The new language must be passed either through a Language instance or an ID
    Args:
        request (HttpRequest): Request object from Django, used to get to the session
        language (Language, optional): A Language instance from this app. Defaults to None.
        language_id (id, optional): ID of the language in our database. Defaults to None.
    """
    # Checking arguments
    if language is None and language_id is None:
        raise TypeError("You must provide either 'language' or 'language_id'")
    # Get the language from the session
    if language is None:
        language = Language.objects.get(id=language_id)
    # Update the user's language
    activate(language.django_language_name)
    request.session[LANGUAGE_SESSION_KEY] = language.django_language_name



# coding: utf-8
"""
Description:
    Contains helper functions and classes for models.Field instances
Fields:
    ForeignKeyCascade: ForeignKey set up for CASCADE 'on delete' with index
    NotEmptyCharField: Charfield that cannot be null nor an empty string
    TranslatableField: Field to use if your field must be translated. It will set a ForeignKey to our "Item" model.
"""


# --------------------------------------------------------------------------------
# > Imports
# --------------------------------------------------------------------------------
# Built-in

# Django
from django.db import models

# Third-party

# Local


# --------------------------------------------------------------------------------
# > Model Fields
# --------------------------------------------------------------------------------
def ForeignKeyCascade(to, *args, **kwargs):
    """ForeignKey set up for CASCADE 'on delete' with index"""
    kwargs['db_index'] = True
    kwargs['on_delete'] = models.CASCADE
    kwargs['null'] = False
    return models.ForeignKey(to, *args, **kwargs)


def NotEmptyCharField(*args, **kwargs):
    """Charfield that cannot be null nor an empty string"""
    kwargs['blank'] = False
    kwargs['null'] = False
    return models.CharField(*args, **kwargs)


def TranslatedField(*args, **kwargs):
    """Field to use if your field must be translated. It will set a ForeignKey to our "Item" model."""
    kwargs['blank'] = True
    kwargs['db_index'] = True
    kwargs['default'] = ""
    kwargs['help_text'] = "The key will be automatically provided when the object is created"
    kwargs['on_delete'] = models.SET_DEFAULT
    kwargs['null'] = True
    return models.ForeignKey("django_database_translation.Item", *args, **kwargs)



# coding: utf8


# --------------------------------------------------------------------------------
# > Imports
# --------------------------------------------------------------------------------
# Built-in
from setuptools import setup

# Third-party

# Local


# --------------------------------------------------------------------------------
# > Functions
# --------------------------------------------------------------------------------
with open("README.md", "r", encoding="utf-8") as f:
    long_description = f.read()

setup(
    # General
    name='django_database_translation',
    version='1.1.3',
    license='MIT',
    # Description
    description='Package to handle database translation in your Django apps',
    long_description=long_description,
    long_description_content_type="text/markdown",
    # Author
    author='Jordan Kowal',
    author_email='kowaljordan@gmail.com',
    # URLs
    url='https://github.com/Jordan-Kowal/django_database_translation',
    download_url='https://github.com/Jordan-Kowal/django_database_translation/archive/v1.1.3.tar.gz',
    # Packages
    packages=['django_database_translation'],
    install_requires=[],
    # Other info
    keywords=["django", "database", "db", "translation", "translate", "backend"],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
    ],
)
